use strict;

use Getopt::Long;
use SeedEnv;
use ExpressionDir;
#
# This is a SAS Component
#


=head1 svr_all_genomes

List the names and IDs of all the (complete) genomes.

There is no input.  The output is a file of genome names and genome IDs.

------
Example: svr_all_genomes -complete > complete_genomes.tbl

would produce a 2-column table.  The first column would contain the
names of all complete genomes, and the second the IDs of those genomes.
------

=head2 Command-Line Options

=over 4

=item url

The URL for the Sapling server, if it is to be different from the default.

=item complete

If TRUE, only complete genomes will be returned. The default is FALSE (return all genomes).

=back

=head2 Output Format

The standard output is a file where each line contains a genome name and a genome ID.

=cut

my $usage    = 'usage: svr_run_gene_activity_simulation --experiment "temperature study,aerobic growth study,pH study" --genome 83333.1 --directory /var/ExpressionData/83333.1/ --model Seed83333.1 --media Carbon-D-Glucose --url http://www.theseed.org/ --user test --password none'."\n";
my $experiment = '';
my $directory = '';
my $genome = '';
my $model = '';
my $media = '';
my $user = '';
my $password = '';
my $url = '';
my $opted    = GetOptions('experiment=s' => \$experiment,'media:s' => \$media,'model:s' => \$model,'genome:s' => \$genome,'directory:s' => \$directory,'user:s' => \$user,'password:s' => \$password,'url:s' => \$url);
if (!$opted) {
    print STDERR $usage;
    return;
}

#Setting default values for variables:
if (length($media) == 0) {
	$media = "Complete";	
}
if (length($experiment) == 0) {
	print STDERR "One or more experiment IDs must be provided for analysis"
	return;
}
if (length($genome) == 0 && length($directory) == 0) {
	print STDERR "Either a genome or experiment directory must be provided!";
	return;
}
if (length($genome) == 0) {
	my $result = ExpressionDir::get_directory_genome_id({directory => $directory});
	if (defined($result->{genomeid})) {
		$genome = $result->{genomeid};
	}
}
my $experimentList;
push(@{$experimentList},split(/,/,$experiment));
my $result = ExpressionDir::get_experiment_on_off_calls({genome => $genome,experiment => $experimentList});
my $fbamodel  = FBAMODELserver->new(url => $url);
#Submitting the jobs
my $jobs;
my $output;
foreach my $exp (keys(%{$result})) {
	my $geneCalls = $result->{$exp};
	my $input = {model => $model,media => $media,geneCalls => $geneCalls};
	if (length($user) > 0) {
		$input->{user} = $user;
		$input->{password} = $password;
	}
	my $result = $fbamodel->fba_submit_gene_activity_analysis($input);
	if (defined($result->{jobid})) {
		$jobs->{$result->{jobid}} = $exp;
	} elsif (defined($result->{jobid})) {
		$output->{$exp} = "FIALED:".$result->{error};	
	} else {
		$output->{$exp} = "FIALED";
	}
}
#Checking on jobs
my $running = 1;
while ($running == 1) {
	sleep(60);
	my $jobList;
	push(@{$jobList},keys(%{$jobs}));
	if (@{$jobList} == 0) {
	 	$running = 0;
	} else {
		my $input = {jobid => $jobList};
		if (length($user) > 0) {
			$input->{user} = $user;
			$input->{password} = $password;
		}
		my $result = $fbamodel->fba_retreive_gene_activity_analysis($input);
		if (defined($result)) {
			if (defined($result->{error})) {
				print STDERR $result->{error};
				return;	
			}
			for (my $i=0; $i < @{$jobList}; $i++) {
				if (!defined($result->{$jobList->[$i]}->{status}) || $result->{$jobList->[$i]}->{status} eq "failed") {
					print STDERR $jobList->[$i]." failed!";
					$output->{$jobs->{$jobList->[$i]}} = "FIALED";
					delete $jobs->{$jobList->[$i]};
				} elsif ($result->{$jobList->[$i]}->{status} eq "complete") {
					$output->{$jobs->{$jobList->[$i]}} = $result->{$jobList->[$i]}->{results};
					ExpressionDir::save_model_gene_activity({expt_name => $result->{$jobList->[$i]}->{results}});
					delete $jobs->{$jobList->[$i]};
				}
			}
		}
	}
}
return $output;